from numbers import Number
import math


# All the imports from pyparsing go here
from typing import Any, List

from types import ModuleType

from pyparsing import (delimitedList, Forward, Literal,
                       stringEnd, nums, Word, CaselessLiteral, Combine,
                       Optional, Suppress, OneOrMore, ZeroOrMore, opAssoc,
                       infixNotation as operatorPrecedence, oneOf, ParseException,
                       ParserElement,
                       alphas, alphanums, ParseFatalException,
                       ParseSyntaxException, FollowedBy, NotAny, Or,
                       MatchFirst, Keyword, Group, White, lineno, col)


# from .pyparsing_utils import myOperatorPrecedence


# Enable memoization (much faster!)
ParserElement.enablePackrat(cache_size_limit=None)
# else:
#     # Pyparsing 2.0
#     from pyparsing import infixNotation
#     myOperatorPrecendence = infixNotation

from .interface import Where


class ParsingTmp:
    # TODO: FIXME: decide on an order, if we do the opposite it doesn't work.
    contract_types: List[Any] = []
    keywords: List[Any] = []


def add_contract(x):
    ParsingTmp.contract_types.append(x)


def add_keyword(x):
    """ Declares that x is a keyword --- this is useful to have more
        clear messages. "keywords" are not parsed by Extension.
        (see extensions.py) and allows to have "deep" error indications.
        See http://pyparsing.wikispaces.com/message/view/home/620225
        and the discussion of the "-" operator in the docs.
    """
    ParsingTmp.keywords.append(x)

W = Where


O = Optional
S = Suppress

basenumber = Word(nums)
point = Literal('.')
e = CaselessLiteral('E')
plusorminus = Literal('+') | Literal('-')
integer = Combine(O(plusorminus) + basenumber)
integer.setParseAction(lambda tokens: SimpleRValue(int(tokens[0])))
floatnumber = Combine(
    O(plusorminus) + integer + (point + O(basenumber)) ^ (e + integer))
floatnumber.setParseAction(lambda tokens: SimpleRValue(float(tokens[0])))
pi = Keyword('pi').setParseAction(
    lambda tokens: SimpleRValue(math.pi, 'pi'))  # @UnusedVariable

np_ = False

try:
    import numpy
except ImportError:
    pass
else:
    np_ = True


def isnumber(x):
    # These are scalar quantities that we can compare (=,>,>=, etc.)
    if isinstance(x, Number):
        return True

    if np_ and isinstance(x, numpy.number):
        return True

    return False

rvalue = Forward()
rvalue.setName('rvalue')
contract_expression = Forward()
contract_expression.setName('contract')
simple_contract = Forward()
simple_contract.setName('simple_contract')

# Import all expressions -- they will call add_contract()
from .library import (EqualTo, Unary, Binary, composite_contract,
                      identifier_contract, misc_variables_contract,
                      scoped_variables_ref,
                      int_variables_contract, int_variables_ref,
                      misc_variables_ref, SimpleRValue)


number = pi | floatnumber | integer
operand = number | int_variables_ref | misc_variables_ref | scoped_variables_ref
operand.setName('r-value')


op = operatorPrecedence
# op  = myOperatorPrecedence
rvalue << op(operand, [
    ('-', 1, opAssoc.RIGHT, Unary.parse_action),
    ('*', 2, opAssoc.LEFT, Binary.parse_action),
    ('-', 2, opAssoc.LEFT, Binary.parse_action),
    ('+', 2, opAssoc.LEFT, Binary.parse_action),
    ('^', 2, opAssoc.LEFT, Binary.parse_action),
])


# I want
# - BindVariable to have precedence to EqualTo(VariableRef)
# but I also want:
# - Arithmetic to have precedence w.r.t BindVariable
# last is variables
add_contract(misc_variables_contract)
add_contract(int_variables_contract)

add_contract(rvalue.copy().setParseAction(EqualTo.parse_action))

hardwired = MatchFirst(ParsingTmp.contract_types)
hardwired.setName('Predefined contract expression')

simple_contract << (hardwired | identifier_contract)
simple_contract.setName('simple contract expression')

any_contract = composite_contract | simple_contract
any_contract.setName('Any simple or composite contract')
contract_expression << (any_contract)  # Parentheses before << !!




# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
from inspect import getouterframes
from contextlib import suppress
from warnings import warn
from typing import Any, AnyStr, Union, Tuple, Dict, Callable

from pyrsistent import v, pvector

from .math import Undefined, Warn


def _apply(f, args=None, kwargs=None):
    return f(*(args or []), **(kwargs or {}))


class Surely:
    """A monad for a successful programmatic unit
    Truthy, defers to an instance of a successful computation"""

    __slots__ = v(
        "__weakref__", "__instance__" "__str__", "__int__", "__float__", "__complex__"
    )

    # def __try_except_undefined__(self, other=None, call=None):
    #     if call:
    #         # noinspection PyComparisonWithNone
    #         return (call(self.__instance__) if other is None else call(self.__instance__, other)
    #         )
    #     else:
    #         return self.__instance__
    #
    # def __hash__(self):
    #     return self.__try_except_undefined__(call=hash)
    #
    # def __eq__(self, other):
    #     return self.__try_except_undefined__(other, op.eq)
    #
    # def __add__(self, other):
    #     return self.__try_except_undefined__(other, op.add)
    #
    # def __sub__(self, other):
    #     return self.__try_except_undefined__(other, op.sub)
    #
    # def __neg__(self):
    #     return self.__try_except_undefined__(call=op.neg)
    #
    # def __invert__(self):
    #     return self.__try_except_undefined__(call=op.invert)
    #
    # def __mul__(self, other):
    #     return self.__try_except_undefined__(other, op.mul)
    #
    # def __truediv__(self, other):
    #     return self.__try_except_undefined__(other, op.truediv)
    #
    # def __floordiv__(self, other):
    #     return self.__try_except_undefined__(other, op.floordiv)
    #
    # def __oct__(self):
    #     return self.__try_except_undefined__(call=oct)
    #
    # def __nonzero__(self):
    #     return self.__try_except_undefined__(call=bool)
    #
    # @classmethod
    # def __call__(cls, *args, **kwargs):
    #     return cls.__instance__

    def __new__(cls, instance__=Undefined(), *args, **kwargs):
        cls.__instance__ = instance__
        # noinspection PyUnresolvedReferences
        return cls.__instance__


# TODO: Refactor so we're not passing around stack frames that may or may not exist.
class Maybe:
    """Sandboxes a Surely call and handles ContractNotRespected by returning Undefined"""

    __slots__ = v("_func", "_args", "_kwargs", "_stack", "_warn", "__instance__")

    def __init__(self, func, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        # noinspection PyUnresolvedReferences,PyProtectedMember
        self._stack = pvector(getouterframes(sys._getframe(0), 1))
        self._warn = Warn(self._stack, self._func, self._args, self._kwargs)
        self.__instance__ = Undefined()

    def __repr__(self) -> str:
        maybe = self.__class__.__name__
        inst = self.__instance__
        inst_undef = inst == Undefined()
        debug = (" -> %s" % Undefined() if inst_undef else " -> %s" % str(inst))
        fname = str(self._func.__name__)
        args = str(self._args).strip("()")
        kwargs = [str(k) + " = " + str(v) for k, v in self._kwargs.items()]
        return maybe + "(" + ", ".join([fname, args, *kwargs]) + ")" + debug

    def __invert__(self) -> Union[Surely, Undefined]:
        return ~self.__instance__

    def __neg__(self) -> Union[Surely, Undefined]:
        return -self.__instance__

    @property
    def func(self) -> Callable:
        return self._func

    @property
    def args(self) -> Tuple[Any]:
        return self._args

    @property
    def kwargs(self) -> Dict[AnyStr, Any]:
        return self._kwargs

    def __call__(self, warn_undefined=True) -> Union[Surely, Undefined]:
        try:
            with suppress(ValueError):
                val = _apply(self.func, self.args, self.kwargs)
            with suppress(UnboundLocalError):
                self.__instance__ = Surely(val)
                # SURELY #
                return self.__instance__
            self.__instance__ = Undefined()
        except Exception as e:
            if warn_undefined:
                w = self._warn.create(e)
                warn(w, category=RuntimeWarning, stacklevel=0)
            # UNDEFINED #
            self.__instance__ = Undefined()
            return self.__instance__




##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

limited_builtins = {}


def limited_range(iFirst, *args):
    # limited range function from Martijn Pieters
    RANGELIMIT = 1000
    if not len(args):
        iStart, iEnd, iStep = 0, iFirst, 1
    elif len(args) == 1:
        iStart, iEnd, iStep = iFirst, args[0], 1
    elif len(args) == 2:
        iStart, iEnd, iStep = iFirst, args[0], args[1]
    else:
        raise AttributeError('range() requires 1-3 int arguments')
    if iStep == 0:
        raise ValueError('zero step for range()')
    iLen = int((iEnd - iStart) / iStep)
    if iLen < 0:
        iLen = 0
    if iLen >= RANGELIMIT:
        raise ValueError(
            'To be created range() object would be to large, '
            'in RestrictedPython we only allow {limit} '
            'elements in a range.'.format(limit=str(RANGELIMIT)),
        )
    return range(iStart, iEnd, iStep)


limited_builtins['range'] = limited_range


def limited_list(seq):
    if isinstance(seq, str):
        raise TypeError('cannot convert string to list')
    return list(seq)


limited_builtins['list'] = limited_list


def limited_tuple(seq):
    if isinstance(seq, str):
        raise TypeError('cannot convert string to tuple')
    return tuple(seq)


limited_builtins['tuple'] = limited_tuple



from Aspidites._vendor.RestrictedPython import compile_restricted_exec


import_errmsg = (
    'Line 1: "%s" is an invalid variable name because it starts with "_"')


def test_RestrictingNodeTransformer__visit_Import__1():
    """It allows importing a module."""
    result = compile_restricted_exec('import a')
    assert result.errors == ()
    assert result.code is not None


def test_RestrictingNodeTransformer__visit_Import__2():
    """It denies importing a module starting with `_`."""
    result = compile_restricted_exec('import _a')
    assert result.errors == (import_errmsg % '_a',)


def test_RestrictingNodeTransformer__visit_Import__3():
    """It denies importing a module starting with `_` as something."""
    result = compile_restricted_exec('import _a as m')
    assert result.errors == (import_errmsg % '_a',)


def test_RestrictingNodeTransformer__visit_Import__4():
    """It denies importing a module as something starting with `_`."""
    result = compile_restricted_exec('import a as _m')
    assert result.errors == (import_errmsg % '_m',)


def test_RestrictingNodeTransformer__visit_Import__5():
    """It allows importing from a module."""
    result = compile_restricted_exec('from a import m')
    assert result.errors == ()
    assert result.code is not None


def test_RestrictingNodeTransformer__visit_Import_6():
    """It allows importing from a module starting with `_`."""
    result = compile_restricted_exec('from _a import m')
    assert result.errors == ()
    assert result.code is not None


def test_RestrictingNodeTransformer__visit_Import__7():
    """It denies importing from a module as something starting with `_`."""
    result = compile_restricted_exec('from a import m as _n')
    assert result.errors == (import_errmsg % '_n',)


def test_RestrictingNodeTransformer__visit_Import__8():
    """It denies as-importing something starting with `_` from a module."""
    result = compile_restricted_exec('from a import _m as n')
    assert result.errors == (import_errmsg % '_m',)


def test_RestrictingNodeTransformer__visit_Import__9():
    """It denies relative from importing as something starting with `_`."""
    result = compile_restricted_exec('from .x import y as _leading_underscore')
    assert result.errors == (import_errmsg % '_leading_underscore',)


def test_RestrictingNodeTransformer__visit_Import_star__1():
    """Importing `*` is a SyntaxError in Python itself."""
    result = compile_restricted_exec('import *')
    assert result.errors == (
        "Line 1: SyntaxError: invalid syntax at statement: 'import *'",)
    assert result.code is None


def test_RestrictingNodeTransformer__visit_Import_star__2():
    """It denies importing `*` from a module."""
    result = compile_restricted_exec('from a import *')
    assert result.errors == ('Line 1: "*" imports are not allowed.',)
    assert result.code is None



from .  import syntax_fail, good, fail

good('dict', {})
fail('dict', 1)
fail('dict', [])
syntax_fail('dict[]')
syntax_fail('dict[]()')
syntax_fail('dict()')
good('dict[1]', {1: 2})
good('dict[N],N<2', {1: 2})
fail('dict[N],N<2', {1: 2, 3: 4})
good('dict(int:int)', {1: 2})
fail('dict(int:int)', {'a': 2})
good('dict(*:int)', {1: 2})
good('dict(*:int)', {'a': 2})

# dictionary of string -> tuple, with tuple of two elements with different type
good('dict(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1.1)})
fail('dict(str:tuple(type(x),type(y))),x!=y', {'a': (2, 1)})



# noinspection PyUnresolvedReferences,PyProtectedMember
from contextlib import _GeneratorContextManager
from typing import Any, Callable, TypeVar

DEF: Any
POS: Any
EMPTY: Any

class FunctionMaker:
    args: Any
    varargs: Any
    varkw: Any
    defaults: Any
    kwonlyargs: Any
    kwonlydefaults: Any
    shortsignature: Any
    name: Any
    doc: Any
    module: Any
    annotations: Any
    signature: Any
    dict: Any
    def __init__(self, func: Any | None = ..., name: Any | None = ..., signature: Any | None = ..., defaults: Any | None = ..., doc: Any | None = ..., module: Any | None = ..., funcdict: Any | None = ...) -> None: ...
    def update(self, func, **kw) -> None: ...
    def make(self, src_templ, evaldict: Any | None = ..., addsource: bool = ..., **attrs): ...
    @classmethod
    def create(cls, obj, body, evaldict, defaults: Any | None = ..., doc: Any | None = ..., module: Any | None = ..., addsource: bool = ..., **attrs): ...

def fix(args, kwargs, sig): ...
F = TypeVar('F', bound=Callable[..., Any])

def decorate(func: F, caller, extras=..., kwsyntax: bool = ...) -> F: ...
def decorator(caller: F, _func: F = ..., kwsyntax: bool = ...) -> F: ...

class ContextManager(_GeneratorContextManager):
    def __init__(self, g, *a, **k): ...
    def __call__(self, func): ...

def contextmanager(func): ...
def append(a, vancestors) -> None: ...
def dispatch_on(*dispatch_args): ...



from Aspidites._vendor.contracts import parse
from Aspidites._vendor.contracts.interface import Where, ContractSyntaxError
from Aspidites._vendor.contracts.library import *  # @UnusedWildImport
from Aspidites._vendor.contracts.syntax import ParseFatalException, ParseException
import unittest


def expression_fails(expression, string, all=True):  # @ReservedAssignment
    try:
        c = expression.parseString(string, parseAll=all)
    except ParseException:
        pass
    except ParseFatalException:
        pass
    else:
        raise Exception('Expression: %s\nparsed to: %s\n(%r)' % 
                        (string, c, c))


def expression_parses(expression, string, all=True):  # @ReservedAssignment
    try:
        expression.parseString(string, parseAll=all)
    except ParseException as e:
        where = Where(string, e.loc)
        msg = 'Error in parsing string: %s' % e
        raise ContractSyntaxError(msg, where=where)
    except ParseFatalException as e:
        where = Where(string, e.loc)
        msg = 'Fatal error in parsing string: %s' % e
        raise ContractSyntaxError(msg, where=where)


class TestParticular(unittest.TestCase):

    def test_variables(self):
        for s in ['a', 'b', 'c', 'd', 'x', 'y']:
            self.assertEqual(parse(s), BindVariable(s, object))
            U = s.upper()
            self.assertEqual(parse(U), BindVariable(U, int))

    def test_variable_parseable(self):
        for s in ['a', 'b', 'c', 'd', 'x', 'y']:
            expression_fails(int_variables_contract, s)
            expression_parses(misc_variables_contract, s)
            U = s.upper()
            expression_parses(int_variables_contract, U)
            expression_fails(misc_variables_contract, U)

    def test_partial(self):
        expression_parses(int_variables_contract, 'A', all=False)
        expression_fails(int_variables_contract, 'A A', all=True)
        expression_parses(int_variables_contract, 'A', all=True)
        expression_fails(int_variables_contract, 'A*', all=False)




class TestBindingVsRef(unittest.TestCase):
    def test_binding_vs_ref(self):
        self.assertEqual(parse('list[N]'), List(BindVariable('N', int), None))

    def test_binding_vs_ref2(self):
        self.assertEqual(parse('N'), BindVariable('N', int))






from Aspidites._vendor.contracts.test_registrar import syntax_fail, good, fail, semantic_fail

from . import dummy_tc
from . import separate_context_tc
from . import types_tc
from . import strings_tc
from . import lists_tc
from . import tuple_tc
from . import dicts_tc
from . import comparison_tc
from . import arithmetic_tc
from . import compositions_tc
from . import variables_tc
from . import simple_values_tc
from . import map_tc
from . import seq_tc
from . import collection_tc
from . import attr_tc
from . import files_tc
from . import miscellaneous_aliases_tc
from . import isinstance_tc
from . import sets_tc
from . import extensions_tc

try:
    import numpy
except ImportError:  # pragma: no cover
    pass
else:
    from . import array_tc
    from . import array_elements_tc




import unittest

from Aspidites._vendor.contracts import new_contract, check, Contract, contract
from Aspidites._vendor.contracts.library.extensions import identifier_expression

from .utils import check_contracts_fail, check_contracts_ok
from Aspidites._vendor.contracts.main import can_be_used_as_a_type, Storage
from Aspidites._vendor.contracts.syntax import ParsingTmp

# The different patterns


def ok1(x):
    pass


def ok2(x):  # @UnusedVariable
    return True


def fail1(x):
    raise ValueError('message')


def fail2(x):  # @UnusedVariable
    return False


def invalid_callable1(x):  # @UnusedVariable
    return 'ciao'


# generates a new name:
def cname():
    TestNewContract.counter += 1
    return 'GeneratedContract%d' % TestNewContract.counter


class TestNewContract(unittest.TestCase):
    counter = 0

    #def setup(self):
    #    Storage.string2contract = {}

    def test_inverted_args(self):
        self.assertRaises(ValueError, new_contract, ok1, 'list')

    def test_wrong_args(self):
        self.assertRaises(ValueError, new_contract, 'my13', 2)

    def test_invalid_callable(self):
        self.assertRaises(ValueError, new_contract, 'new', lambda: None)

    def test_parsing_error(self):
        self.assertRaises(ValueError, new_contract, 'new', '>>')

    def test_parsing_error2(self):
        # parsing error (unknown spec)
        self.assertRaises(ValueError, new_contract, 'new', 'unknown')

    def test_invalid_names(self):
        # invalid names:
        alphabet = 'A B C D E F G H I J K L M N O P Q R S T U W V X Y Z'
        for x in alphabet.split():
            self.assertRaises(ValueError, new_contract, x, 'list')
            self.assertRaises(ValueError, new_contract, x.lower(), 'list')
        self.assertRaises(ValueError, new_contract, 'list', 'list[N]')
        self.assertRaises(ValueError, new_contract, '2acdca', 'list[N]')
        self.assertRaises(ValueError, new_contract, '_', 'list[N]')

    def test_valid(self):
        c = new_contract('my_list', 'list[2]')
        assert isinstance(c, Contract)
        check('tuple(my_list, my_list)', ([1, 2], [1, 2]))
        check_contracts_fail('tuple(my_list, my_list)', ([1, 2], [1, 2, 3]))

    def test_separate_context(self):
        new_contract('my_list2', 'list[N]')
        check('tuple(my_list2, my_list2)', ([1, 2], [1, 2]))
        check('tuple(my_list2, my_list2)', ([1, 2], [1, 2, 3]))

    def test_renaming(self):
        self.assertNotEqual(ok1, ok2)
        new_contract('my7', ok1)
        self.assertRaises(ValueError, new_contract, 'my7', ok2)

    def test_allow_renaming_if_equal1(self):
        new_contract('my8', ok1)
        new_contract('my8', ok1)

    def test_allow_renaming_if_equal2(self):
        new_contract('my8b', 'list[3]')
        new_contract('my8b', 'list[3]')

    def test_callable1(self):
        c = cname()
        new_contract(c, ok2)
        check('list(%s)' % c, [0])

    def test_callable2(self):
        c = cname()
        new_contract(c, ok2)
        check('list(%s)' % c, [0])

    def test_callable3(self):
        c = cname()
        new_contract(c, fail1)
        check_contracts_fail('list(%s)' % c, [0])

    def test_callable4(self):
        c = cname()
        new_contract(c, fail2)
        check_contracts_fail('list(%s)' % c, [0])

    def test_invalid_callable2(self):
        c = cname()
        new_contract(c, invalid_callable1)
        self.assertRaises(ValueError, check, 'list(%s)' % c, [0])

    def test_other_pass(self):
        class Ex1(Exception):
            pass

        def invalid(x):
            raise Ex1()
        c = cname()
        new_contract(c, invalid)
        self.assertRaises(Ex1, check, 'list(%s)' % c, [0])

    def test_callable(self):
        class MyTest_ok(object):
            def __call__(self, x):  # @UnusedVariable
                return True
        o = MyTest_ok()
        assert o('value') == True
        new_contract(cname(), o)

    def test_callable_old_style(self):
        class MyTest_ok():
            def __call__(self, x):  # @UnusedVariable
                return True
        o = MyTest_ok()
        assert o('value') == True
        new_contract(cname(), o)

    def test_class_as_contract1(self):
        # This should be interpreted as a type
        # init(x,y) so it is not mistaken for a valid callable
        class NewStyleClass(object):
            def __init__(self, x, y):  # @UnusedVariable
                pass
        new_contract(cname(), NewStyleClass)

    def test_class_as_contract2(self):
        # old sytle class
        class OldStyleClass():
            def __init__(self, x, y):  # @UnusedVariable
                pass
        new_contract(cname(), OldStyleClass)

    def test_class_as_contract3(self):
        class NewStyleClass(object):
            def __init__(self, x, y):  # @UnusedVariable
                pass

        @contract(x=NewStyleClass)
        def f(x):
            pass

    def test_class_as_contract4(self):
        class OldStyleClass():
            def __init__(self, x, y):  # @UnusedVariable
                pass

        @contract(x=OldStyleClass)
        def f(x):
            pass

    def test_callable_5(self):
        class MyTest_ok(object):
            def f(self, x):  # @UnusedVariable
                return True
        o = MyTest_ok()
        assert o.f('value') == True
        new_contract(cname(), o.f)

# TODO: removed after checking in class methods
#    def test_callable_invalid(self):
#        class MyTest_fail(object):
#            def __call__(self, x, y):  # @UnusedVariable
#                return True
#
#        self.assertRaises(ValueError, new_contract, cname(), MyTest_fail())

    def test_lambda_2(self):
        new_contract(cname(), lambda x: True)  # @UnusedVariable
        new_contract(cname(), lambda x: None)  # @UnusedVariable

    def test_lambda_invalid(self):
        f = lambda: True  # @UnusedVariable
        self.assertRaises(ValueError, new_contract, cname(), f)

    def test_idioms(self):
        Storage.string2contract = {}  # TODO remove this hack
        color = new_contract('color', 'list[3](number,>=0,<=1)')
        # Make sure we got it right
        color.check([0, 0, 0])
        color.check([0, 0, 1])
        color.fail([0, 0])
        color.fail([0, 0, 2])

        self.assertRaises(ValueError, color.fail, [0, 0, 1])

        # Now use ``color`` in other contracts.
        @contract
        def fill_area(inside, border):
            """ Fill the area inside the current figure.

                :type border: color
                :type inside: color
            """
            pass

        @contract
        def fill_gradient(colors):
            """ Use a gradient to fill the area.

                :type colors: list(color)
            """
            pass

    def test_as_decorator(self):
        @new_contract
        def even(x):
            return x % 2 == 0
        from Aspidites._vendor.contracts import parse
        p = parse('even')
        p.check(2)
        p.check(4)
        p.fail(3)
        p.check(2.0)

    def test_as_decorator_multiple(self):
        @new_contract
        @contract(x='int')
        def even2(x):
            return x % 2 == 0

        from Aspidites._vendor.contracts import parse
        p = parse('even2')
        p.check(2)
        p.fail(3)
        p.fail(2.0) # now fails

    def test_types_as_contracts(self):
        c = cname()
        new_contract(c, str)
        check_contracts_ok(c, '')
        check_contracts_fail(c, 1)

    def test_types_as_contracts2(self):
        c = cname()
        new_contract(c, int)
        check_contracts_ok(c, 1)
        check_contracts_fail(c, '')

    def test_well_recognized(self):
        class OldStyleClass():
            def __init__(self, x, y):  # @UnusedVariable
                pass

        assert can_be_used_as_a_type(OldStyleClass)

        class NewStyleClass():
            def __init__(self, x, y):  # @UnusedVariable
                pass

        assert can_be_used_as_a_type(NewStyleClass)

    def test_as_decorator_with_args(self):
        from Aspidites._vendor.contracts import parse

        @new_contract
        def greater_than(value, thresh):
            return value > thresh

        p = parse('greater_than(5)')
        p.check(6)
        p.fail(5)

    def test_as_decorator_with_kwargs(self):
        from Aspidites._vendor.contracts import parse

        @new_contract
        def less_than(value, thresh=1):
            return value < thresh

        p = parse('less_than(thresh=2)')
        p.check(1)
        p.fail(2)

    def test_as_decorator_with_args_and_kwargs(self):
        from Aspidites._vendor.contracts import parse

        @new_contract
        def less_than_all(value, a, b, c=1, d=5):
            return value < min(a, b, c, d)

        p = parse('less_than_all(3, 4, c=5)')
        p.check(2)
        p.fail(3)

    def test_capital_name1(self):
        # some problems with identifiers starting with capitals
        new_contract('SEn', 'int')
        check_contracts_ok('SEn', 1)
        check_contracts_fail('SEn', 2.0)

    def test_capital_name2(self):
        new_contract('Sen', 'int')
        check_contracts_ok('Sen', 1)
        check_contracts_fail('Sen', 2.0)


examples_valid = ['aa', 'a_', 'a2', 'a_2', 'list2', 'dict2', 'int2',
                'float2', 'point2', 'A2', 'array2', 'unit_length', 'SE2', 'SE3', 'S1', 'S2', 'axis_angle']

for k in ParsingTmp.keywords:
    examples_valid.append('%s2' % k)
    examples_valid.append('%s_' % k)


def check_valid_identifier(e):
    check_valid_identifier.__dict__['description'] = \
        'check_valid_identifier(%r)' % e

    identifier_expression.parseString(e, parseAll=True) #@UndefinedVariable

    new_contract(e, '*')

    check(e, 42)


# def test_valid_identifiers():
#
#     for e in examples_valid:
#         yield check_valid_identifier, e





import sys
from warnings import warn
from typing import Any, Union
from inspect import isfunction, signature, getouterframes
from cmath import inf, isinf, nan, isnan
import numbers
import numpy as np
from pyrsistent import v, pvector
from .templates import _warning
from .api import bordered

Numeric = Union[int, float, complex, np.number]


class Undefined:
    """A monad for a failed programmatic unit; like NoneType but hashable.
    Falsy singleton acts as an absorbing element for division."""

    __slots__ = v("__weakref__", "__instance__")
    __instance = None

    def __hash__(self):
        # noinspection PyUnresolvedReferences
        return hash(self.__weakref__)

    def __eq__(self, other: Any):
        return self.__hash__ == other.__hash__

    def __add__(self, other: Any):
        return self

    def __sub__(self, other: Any):
        return self

    def __mul__(self, other: Any):
        return self

    def __truediv__(self, other: Any):
        return self

    def __floordiv__(self, other):
        return self

    def __neg__(self):
        return self

    def __invert__(self):
        return self

    def __str__(self):
        return self.__repr__()

    def __float__(self):
        return nan

    def __complex__(self):
        return complex(nan)

    def __oct__(self):
        return self

    def __index__(self):
        return 0

    def __len__(self):
        # Undefined has 0 elements
        return 0

    def __repr__(self):
        return self.__class__.__name__ + "()"

    # noinspection PyMethodMayBeStatic
    def __nonzero__(self):
        return True

    def __call__(self, *args, **kwargs):
        return self.__new__(self.__class__, *args, **kwargs)

    # noinspection PyMethodParameters
    def __new__(mcs, *args, **kwargs):
        if mcs.__instance is None:
            mcs.__instance = super(Undefined, mcs).__new__(mcs, *args, **kwargs)
            mcs.__instance__ = mcs.__instance
        # noinspection PyUnresolvedReferences
        return mcs.__instance__  # instance descriptor from __slots__ -> actual instance


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnaryAdd(a: Numeric) -> Union[Numeric, Undefined]:
    if isnan(a) or not isinstance(a, numbers.Number):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Unary Add is Undefined for %s" % type(a))
        w = Warn(stack, stack[0][3], [a], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return +a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeUnarySub(a: Numeric) -> Union[Numeric, Undefined]:
    if isnan(a) or not isinstance(a, numbers.Number):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Unary Sub is Undefined for %s" % type(a))
        w = Warn(stack, stack[0][3], [a], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return -a


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeFloorDiv(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if isinf(a) or b == 0 or (isinf(a) and isinf(b)):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return a // b


# noinspection PyPep8Naming,PyProtectedMember,PyUnresolvedReferences
def SafeDiv(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if b == 0 or (isinf(a) and isinf(b)):
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Division by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return a / b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeMod(a: Numeric, b: Numeric) -> Union[Numeric, Undefined]:
    """IEEE 754-1985 evaluates an expression and replaces indeterminate forms with Undefined instances"""
    if isinf(a) or b == 0:
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ZeroDivisionError("Modulus by zero is Undefined; this behavior diverges from IEEE 754-1985.")
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    return a % b


# noinspection PyPep8Naming, PyProtectedMember,PyUnresolvedReferences
def SafeExp(a, b):
    if (a == 0 and b == 0) or (isinf(a) and b == 0) or (isinf(b) and a == 0):  # 0**0, inf**0, 0**inf
        stack = pvector(getouterframes(sys._getframe(0), 1))
        exc = ArithmeticError(
            "%s**%s" % (str(a), str(b),) + " == Undefined; this behavior diverges from IEEE 754-1985."
        )
        w = Warn(stack, stack[0][3], [a, b], {}).create(exc)
        warn(w, category=RuntimeWarning, stacklevel=0)
        return Undefined()
    try:
        return a**b
    except OverflowError:
        return inf  # just a really big number on most systems


class Warn:

    def __init__(self, stack, func, *args, **kwargs):
        self.stack = stack
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def create(self, exc=Exception()):
        _locals = self.stack[1][0].f_locals.items()
        str_locals = self.format_locals(_locals, exc)
        func_name = self.stack[1][0].f_code.co_name
        fname = self.stack[1][0].f_code.co_filename
        lineno = self.stack[1][0].f_code.co_firstlineno
        fkwargs = self.format_kwargs()
        if hasattr(self.func, "__name__"):
            name = self.func.__name__
        else:
            name = str(self.func)
        atfault = (
            name
            if isinstance(exc, TypeError)
            else name + "(" + str(self.args).strip("()") + fkwargs + ")"
        )
        return _warning.safe_substitute(
            file=fname,
            lineno=lineno,
            func=bordered(func_name),
            atfault=bordered(atfault),
            bound=bordered(str_locals),
            tb=bordered(str(exc)),
        )

    def format_kwargs(self, sep: str = ", "):
        return sep + str(self.kwargs).strip("{} ").replace(":", "=") if len(self.kwargs) else ""

    # noinspection PyMethodMayBeStatic
    def format_locals(self, local_vars, exc: Exception):
        locals_ = dict(filter(lambda x: x[1] != str(exc), local_vars))
        str_locals = str()
        for k, v_ in locals_.items():
            if str(k).startswith("@"):  # skip @py_assert
                continue
            if isfunction(v_):
                str_locals += k + ": " + str(signature(v_)).replace("'", "") + "\n"
            else:
                str_locals += k + ": " + str(v_) + "\n"
        return str_locals.rstrip("\n")




import unittest

from Aspidites._vendor.contracts import (decorate, contract,
                                         ContractException, ContractNotRespected)

from Aspidites._vendor.contracts.interface import MissingContract

import sys
import doctest
import unittest
import decimal
import inspect
from asyncio import get_event_loop
from collections import defaultdict, ChainMap, abc as c
from Aspidites._vendor.decorator import dispatch_on, contextmanager, decorator
# try:
#     from . import decorator_documentation as doc  # good with pytest
# except ImportError:
#     import decorator_documentation as doc  # good with `python src/tests/test.py`


@contextmanager
def assertRaises(etype):
    """This works in Python 2.6 too"""
    try:
        yield
    except etype:
        pass
    else:
        raise Exception('Expected %s' % etype.__name__)


@decorator
async def before_after(coro, *args, **kwargs):
    return "<before>" + (await coro(*args, **kwargs)) + "<after>"


@decorator
def coro_to_func(coro, *args, **kw):
    return get_event_loop().run_until_complete(coro(*args, **kw))


class CoroutineTestCase(unittest.TestCase):
    def test_before_after(self):
        @before_after
        async def coro(x):
            return x
        self.assertTrue(inspect.iscoroutinefunction(coro))
        out = get_event_loop().run_until_complete(coro('x'))
        self.assertEqual(out, '<before>x<after>')

    def test_coro_to_func(self):
        @coro_to_func
        async def coro(x):
            return x
        self.assertFalse(inspect.iscoroutinefunction(coro))
        self.assertEqual(coro('x'), 'x')


def gen123():
    yield 1
    yield 2
    yield 3


class GeneratorCallerTestCase(unittest.TestCase):
    def test_gen123(self):
        @decorator
        def square(func, *args, **kw):
            for x in gen123():
                yield x * x
        new = square(gen123)
        self.assertTrue(inspect.isgeneratorfunction(new))
        self.assertEqual(list(new()), [1, 4, 9])


# class DocumentationTestCase(unittest.TestCase):
#     def test(self):
#         err = doctest.testmod(doc)[0]
#         self.assertEqual(err, 0)
#
#     def test_copy_dunder_attrs(self):
#         traced = doc.trace(doc.foo)
#         self.assertIn('documentation', traced.__module__)
#         self.assertEqual(traced.__annotations__, {})
#         self.assertEqual(traced.__defaults__, (None,))
#
#     def test_singledispatch1(self):
#         with assertRaises(RuntimeError):
#             doc.singledispatch_example1()
#
#     def test_singledispatch2(self):
#         doc.singledispatch_example2()
#
#     def test_context_manager(self):
#
#         @contextmanager
#         def before_after(before, after):
#             print(before)
#             yield
#             print(after)
#
#         @before_after('BEFORE', 'AFTER')
#         def hello_user(user):
#             print('hello %s' % user)
#
#         argspec = inspect.getfullargspec(hello_user)
#         self.assertEqual(argspec.args, ['user'])
#
#
# class ExtraTestCase(unittest.TestCase):
#     def test_qualname(self):
#         self.assertEqual(doc.operation1.__qualname__, 'operation1')
#
#     def test_signature(self):
#         sig = inspect.signature(doc.f1)
#         self.assertEqual(str(sig), '(x)')
#
#     def test_unique_filenames(self):
#         @decorator
#         def d1(f, *args, **kwargs):
#             return f(*args, **kwargs)
#
#         @decorator
#         def d2(f, *args, **kwargs):
#             return f(*args, **kwargs)
#
#         @d1
#         def f1(x, y, z):
#             pass
#
#         @d2
#         def f2(x, y, z):
#             pass
#
#         f1_orig = f1
#
#         @d1
#         def f1(x, y, z):
#             pass
#
#         self.assertEqual(d1.__code__.co_filename,
#                          d2.__code__.co_filename)
#         self.assertEqual(f1.__code__.co_filename,
#                          f2.__code__.co_filename)
#         self.assertEqual(f1_orig.__code__.co_filename,
#                          f1.__code__.co_filename)
#
#     def test_no_first_arg(self):
#         @decorator
#         def example(*args, **kw):
#             return args[0](*args[1:], **kw)
#
#         @example
#         def func(**kw):
#             "Docstring"
#             return kw
#
#         # there is no confusion when passing args as a keyword argument
#         self.assertEqual(func(args='a'), {'args': 'a'})
#         self.assertEqual(func.__doc__, "Docstring")
#
#     def test_decorator_factory(self):
#         # similar to what IPython is doing in traitlets.config.application
#         @decorator
#         def catch_config_error(method, app, *args, **kwargs):
#             return method(app)
#         catch_config_error(lambda app, **kw: None)(1)
#
#     def test_add1(self):
#         # similar to what IPython is doing in traitlets.config.application
#         @decorator
#         def add(func, const=1, *args, **kwargs):
#             return const + func(*args, **kwargs)
#
#         def f(x):
#             return x
#         self.assertEqual(add(f, 2)(0), 2)
#
#     def test_dan_schult(self):
#         # see https://github.com/micheles/decorator/issues/120
#         @decorator
#         def prnt(func, index=0, *args, **kw):
#             # print(args[index])
#             return func(*args, **kw)
#
#         @prnt(index=2)  # print the value of the third argument
#         def f(a, b, c=None):
#             return [a, b, c]
#
#         self.assertEqual(f(0, 1), [0, 1, None])
#
#     def test_slow_wrapper(self):
#         # see https://github.com/micheles/decorator/issues/123
#         dd = defaultdict(list)
#         doc.trace(defaultdict.__setitem__)(dd, 'x', [1])
#         self.assertEqual(dd['x'], [1])
#         doc.trace(defaultdict.__delitem__)(dd, 'x')
#         self.assertEqual(dd['x'], [])
#         # NB: defaultdict.__getitem__ has no signature and cannot be
#         # decorated in CPython, while it is regular in PyPy


# ################### test dispatch_on ############################# #
# adapted from test_functools in Python 3.5
singledispatch = dispatch_on('obj')


class TestSingleDispatch(unittest.TestCase):
    def test_simple_overloads(self):
        @singledispatch
        def g(obj):
            return "base"

        @g.register(int)
        def g_int(i):
            return "integer"

        self.assertEqual(g("str"), "base")
        self.assertEqual(g(1), "integer")
        self.assertEqual(g([1, 2, 3]), "base")

    def test_mro(self):
        @singledispatch
        def g(obj):
            return "base"

        class A(object):
            pass

        class C(A):
            pass

        class B(A):
            pass

        class D(C, B):
            pass

        @g.register(A)
        def g_A(a):
            return "A"

        @g.register(B)
        def g_B(b):
            return "B"

        self.assertEqual(g(A()), "A")
        self.assertEqual(g(B()), "B")
        self.assertEqual(g(C()), "A")
        self.assertEqual(g(D()), "B")

    def test_register_decorator(self):
        @singledispatch
        def g(obj):
            return "base"

        @g.register(int)
        def g_int(i):
            return "int %s" % (i,)
        self.assertEqual(g(""), "base")
        self.assertEqual(g(12), "int 12")

    def test_register_error(self):
        @singledispatch
        def g(obj):
            return "base"

        with assertRaises(TypeError):
            # wrong number of arguments
            @g.register(int)
            def g_int():
                return "int"

    def test_wrapping_attributes(self):
        @singledispatch
        def g(obj):
            "Simple test"
            return "Test"
        self.assertEqual(g.__name__, "g")
        if sys.flags.optimize < 2:
            self.assertEqual(g.__doc__, "Simple test")

    def test_c_classes(self):
        @singledispatch
        def g(obj):
            return "base"

        @g.register(decimal.DecimalException)
        def _(obj):
            return obj.args
        subn = decimal.Subnormal("Exponent < Emin")
        rnd = decimal.Rounded("Number got rounded")
        self.assertEqual(g(subn), ("Exponent < Emin",))
        self.assertEqual(g(rnd), ("Number got rounded",))

        @g.register(decimal.Subnormal)
        def _g(obj):
            return "Too small to care."
        self.assertEqual(g(subn), "Too small to care.")
        self.assertEqual(g(rnd), ("Number got rounded",))

    def test_register_abc(self):
        d = {"a": "b"}
        l = [1, 2, 3]
        s = set([object(), None])
        f = frozenset(s)
        t = (1, 2, 3)

        @singledispatch
        def g(obj):
            return "base"

        self.assertEqual(g(d), "base")
        self.assertEqual(g(l), "base")
        self.assertEqual(g(s), "base")
        self.assertEqual(g(f), "base")
        self.assertEqual(g(t), "base")

        g.register(c.Sized)(lambda obj: "sized")
        self.assertEqual(g(d), "sized")
        self.assertEqual(g(l), "sized")
        self.assertEqual(g(s), "sized")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sized")

        g.register(c.MutableMapping)(lambda obj: "mutablemapping")
        self.assertEqual(g(d), "mutablemapping")
        self.assertEqual(g(l), "sized")
        self.assertEqual(g(s), "sized")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sized")

        g.register(ChainMap)(lambda obj: "chainmap")
        # irrelevant ABCs registered
        self.assertEqual(g(d), "mutablemapping")
        self.assertEqual(g(l), "sized")
        self.assertEqual(g(s), "sized")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sized")

        g.register(c.MutableSequence)(lambda obj: "mutablesequence")
        self.assertEqual(g(d), "mutablemapping")
        self.assertEqual(g(l), "mutablesequence")
        self.assertEqual(g(s), "sized")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sized")

        g.register(c.MutableSet)(lambda obj: "mutableset")
        self.assertEqual(g(d), "mutablemapping")
        self.assertEqual(g(l), "mutablesequence")
        self.assertEqual(g(s), "mutableset")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sized")

        g.register(c.Mapping)(lambda obj: "mapping")
        self.assertEqual(g(d), "mutablemapping")  # not specific enough
        self.assertEqual(g(l), "mutablesequence")
        self.assertEqual(g(s), "mutableset")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sized")

        g.register(c.Sequence)(lambda obj: "sequence")
        self.assertEqual(g(d), "mutablemapping")
        self.assertEqual(g(l), "mutablesequence")
        self.assertEqual(g(s), "mutableset")
        self.assertEqual(g(f), "sized")
        self.assertEqual(g(t), "sequence")

        g.register(c.Set)(lambda obj: "set")
        self.assertEqual(g(d), "mutablemapping")
        self.assertEqual(g(l), "mutablesequence")
        self.assertEqual(g(s), "mutableset")
        self.assertEqual(g(f), "set")
        self.assertEqual(g(t), "sequence")

        g.register(dict)(lambda obj: "dict")
        self.assertEqual(g(d), "dict")
        self.assertEqual(g(l), "mutablesequence")
        self.assertEqual(g(s), "mutableset")
        self.assertEqual(g(f), "set")
        self.assertEqual(g(t), "sequence")

        g.register(list)(lambda obj: "list")
        self.assertEqual(g(d), "dict")
        self.assertEqual(g(l), "list")
        self.assertEqual(g(s), "mutableset")
        self.assertEqual(g(f), "set")
        self.assertEqual(g(t), "sequence")

        g.register(set)(lambda obj: "concrete-set")
        self.assertEqual(g(d), "dict")
        self.assertEqual(g(l), "list")
        self.assertEqual(g(s), "concrete-set")
        self.assertEqual(g(f), "set")
        self.assertEqual(g(t), "sequence")

        g.register(frozenset)(lambda obj: "frozen-set")
        self.assertEqual(g(d), "dict")
        self.assertEqual(g(l), "list")
        self.assertEqual(g(s), "concrete-set")
        self.assertEqual(g(f), "frozen-set")
        self.assertEqual(g(t), "sequence")

        g.register(tuple)(lambda obj: "tuple")
        self.assertEqual(g(d), "dict")
        self.assertEqual(g(l), "list")
        self.assertEqual(g(s), "concrete-set")
        self.assertEqual(g(f), "frozen-set")
        self.assertEqual(g(t), "tuple")

    def test_mro_conflicts(self):
        @singledispatch
        def g(obj):
            return "base"

        class O(c.Sized):
            def __len__(self):
                return 0
        o = O()
        self.assertEqual(g(o), "base")
        g.register(c.Iterable)(lambda arg: "iterable")
        g.register(c.Container)(lambda arg: "container")
        g.register(c.Sized)(lambda arg: "sized")
        g.register(c.Set)(lambda arg: "set")
        self.assertEqual(g(o), "sized")
        c.Iterable.register(O)
        self.assertEqual(g(o), "sized")
        c.Container.register(O)
        with assertRaises(RuntimeError):  # was "sized" because in mro
            self.assertEqual(g(o), "sized")
        c.Set.register(O)
        self.assertEqual(g(o), "set")

        class P(object):
            pass
        p = P()
        self.assertEqual(g(p), "base")
        c.Iterable.register(P)
        self.assertEqual(g(p), "iterable")
        c.Container.register(P)

        with assertRaises(RuntimeError):
            self.assertEqual(g(p), "iterable")

        class Q(c.Sized):
            def __len__(self):
                return 0
        q = Q()
        self.assertEqual(g(q), "sized")
        c.Iterable.register(Q)
        self.assertEqual(g(q), "sized")
        c.Set.register(Q)
        self.assertEqual(g(q), "set")
        # because c.Set is a subclass of c.Sized and c.Iterable

        @singledispatch
        def h(obj):
            return "base"

        @h.register(c.Sized)
        def h_sized(arg):
            return "sized"

        @h.register(c.Container)
        def h_container(arg):
            return "container"
        # Even though Sized and Container are explicit bases of MutableMapping,
        # this ABC is implicitly registered on defaultdict which makes all of
        # MutableMapping's bases implicit as well from defaultdict's
        # perspective.
        with assertRaises(RuntimeError):
            self.assertEqual(h(defaultdict(lambda: 0)), "sized")

        class R(defaultdict):
            pass
        c.MutableSequence.register(R)

        @singledispatch
        def i(obj):
            return "base"

        @i.register(c.MutableMapping)
        def i_mapping(arg):
            return "mapping"

        @i.register(c.MutableSequence)
        def i_sequence(arg):
            return "sequence"
        r = R()
        with assertRaises(RuntimeError):  # was no error
            self.assertEqual(i(r), "sequence")

        class S(object):
            pass

        class T(S, c.Sized):
            def __len__(self):
                return 0
        t = T()
        self.assertEqual(h(t), "sized")
        c.Container.register(T)
        self.assertEqual(h(t), "sized")   # because it's explicitly in the MRO

        class U(object):
            def __len__(self):
                return 0
        u = U()
        self.assertEqual(h(u), "sized")
        # implicit Sized subclass inferred
        # from the existence of __len__()

        c.Container.register(U)
        # There is no preference for registered versus inferred ABCs.
        with assertRaises(RuntimeError):
            h(u)


class DecoratorTests(unittest.TestCase):

    def test_malformed(self):
        def f():
            """
                Wrong syntax

                :rtype okok
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_malformed2(self):
        def f():
            """
                Wrong syntax

                :rtype: okok
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_not_specified1(self):
        """ No docstring specified """
        def f():
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_not_specified2(self):
        def f():
            """ No types specified in the docstring """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_too_many(self):
        def f():
            """
                Too many rtype clauses.
                :rtype: int
                :rtype: int
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_invalid1(self):
        def f(a):
            """ Unknown b.
                :type a: int
                :type b: int
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_parse_error1(self):
        def f(a, b):
            """ Same with optional
                :type a: in
                :type b: int
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_parse_error2(self):
        def f(a, b):
            """ Same with optional
                :type a: int
                :type b: int
                :rtype: in
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def not_supported1(self):
        """ Support of *args """

        def f(a, *b):  # @UnusedVariable
            """
                :type a: int
                :type b: tuple(int)
                :rtype: int
            """
            pass

            decorate(f)

    def not_supported2(self):
        """ Support of **args """
        def f(a, **b):
            """
                :type a: int
                :type b: dict(int:int)
                :rtype: int
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_ok1(self):
        @contract
        def f(a, b):
            """ This is good
                :type a: int
                :type b: int
                :rtype: int
            """
            pass

    def test_ok3(self):
        """ Trying the quoting thing. """
        @contract
        def f(a, b):
            """ This is good
                :type a: ``int``
                :type b: ``int``
                :rtype: ``int``
            """
            pass

    def test_bad_quoting(self):
        def f(a, b):
            """
                :type a: ``int``
                :type b: ``int``
                :rtype: ``int`
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_bad_quoting2(self):
        def f(a, b):
            """
                :type a: ``int``
                :type b: `int``
                :rtype: ``int``
            """
            pass

        self.assertRaises(ContractException, decorate, f)

    def test_ok2(self):
        @contract(a='int', returns='int')
        def f(a, b):
            pass

    def test_invalid_args(self):
        def f():
            @contract(1)
            def g(a, b):
                return int(a + b)
        self.assertRaises(ContractException, f)

    def test_invalid_args2(self):
        """ unknown parameter """
        def f():
            @contract(c=2)
            def g(a, b):
                return int(a + b)
        self.assertRaises(ContractException, f)

    def test_check_it_works1(self):
        @contract(a='int', b='int', returns='int')
        def f(a, b):  # @UnusedVariable
            return 2.0
        self.assertRaises(ContractNotRespected, f, 1, 2)

    def test_check_it_works2(self):
        @contract(a='int', b='int', returns='int')
        def f(a, b):  # @UnusedVariable
            return a + b
        f(1, 2)
        self.assertRaises(ContractNotRespected, f, 1.0, 2)
        self.assertRaises(ContractNotRespected, f, 1, 2.0)

    def test_check_it_works2b(self):
        """ Nothing for b """
        @contract(a='int', returns='int')
        def f(a, b):  # @UnusedVariable
            return int(a + b)
        f(1, 2)
        f(1, 2.0)

    def test_check_it_works2c(self):
        """ Nothing for b """
        def f1(a, b):  # @UnusedVariable
            return int(a + b)

        f = decorate(f1, a='int', returns='int')

        f(1, 2)
        f(1, 2.0)
        self.assertRaises(ContractNotRespected, f, 1.0, 2)

    # def test_module_as_decorator(self):
    #     import contracts as contract_module
    #
    #     @contract_module
    #     def f(a, b): #@UnusedVariable
    #         return a + b
    #     f(1, 2)
    #     self.assertRaises(ContractNotRespected, f, 1.0, 2)

    def test_check_it_works3(self):
        @contract
        def f(a, b):
            """ This is good
                :type a: int
                :type b: int
                :rtype: int
            """
            return a + b
        f(1, 2)
        self.assertRaises(ContractNotRespected, f, 1.0, 2)
        self.assertRaises(ContractNotRespected, f, 1, 2.0)

    def test_inline_docstring_format_works(self):
        @contract
        def f(a, b):
            """ This is good
                :param int,>0 a: Description
                :param int,>0 b: Description
                :returns int,>0: Description
            """
            return a + b
        f(1, 2)
        self.assertRaises(ContractNotRespected, f, 1.0, 2)
        self.assertRaises(ContractNotRespected, f, -1, 2)

    def test_check_docstring_maintained(self):
        def f1(a, b):
            """ This is good
                :type a: int
                :type b: int
                :rtype: int
            """
            return a + b

        def f2(string):
            pass

        f1_dec = decorate(f1)
        self.assertNotEqual(f1.__doc__, f1_dec.__doc__)
        self.assertEqual(f1.__name__, f1_dec.__name__)
        self.assertEqual(f1.__module__, f1_dec.__module__)

        f2_dec = decorate(f2, string='str')
        self.assertNotEqual(f2.__doc__, f2_dec.__doc__)
        self.assertEqual(f2.__name__, f2_dec.__name__)
        self.assertEqual(f2.__module__, f2_dec.__module__)

        f1_dec_p = decorate(f1, modify_docstring=False)
        self.assertEqual(f1_dec_p.__doc__, f1.__doc__)

        f2_dec_p = decorate(f2, modify_docstring=False, string='str')
        self.assertEqual(f2.__doc__, f2_dec_p.__doc__)

        @contract
        def f1b(a, b):
            """ This is good
                :type a: int
                :type b: int
                :rtype: int
            """
            return a + b

        @contract(string='str')
        def f2b(string):
            pass

        @contract(modify_docstring=False)
        def f1b_p(a, b):
            """ This is good
                :type a: int
                :type b: int
                :rtype: int
            """
            return a + b

        @contract(modify_docstring=False, string='str')
        def f2b_p(string):
            pass

        self.assertNotEqual(f1.__doc__, f1b.__doc__)
        self.assertEqual(f1.__doc__, f1b_p.__doc__)
        self.assertNotEqual(f2.__doc__, f2b.__doc__)
        self.assertEqual(f2.__doc__, f2b_p.__doc__)

    def test_kwargs(self):
        def f(a, b, c=7):  # @UnusedVariable
            """ Same with optional
                :type a: int
                :type b: int
                :type c: int
            """
            if c != b:
                raise Exception()

        f2 = decorate(f)
        f2(0, 7)
        f2(0, 5, 5)
        self.assertRaises(Exception, f2, 0, 5, 4)
        self.assertRaises(Exception, f2, 0, 5)

    def test_varargs(self):
        def f(a, b, *c):
            """ Same with optional
                :type a: int
                :type b: int
                :type c: tuple
            """
            assert c == (a, b)

        f2 = decorate(f)
        f2(0, 7, 0, 7)

    def test_keywords(self):
        def f(A, B, **c):
            """ Same with optional
                :type A: int
                :type B: int
                :type c: dict
            """
            assert c['a'] == A
            assert c['b'] == B

        f2 = decorate(f)
        f(0, 7, a=0, b=7)
        f2(0, 7, a=0, b=7)

        self.assertRaises(Exception, f2, 0, 5, 0, 6)

    def test_same_signature(self):
        from inspect import getfullargspec

        def f(a):
            return a

        @contract(a='int')
        def f2(a):
            return a

        self.assertEqual(getfullargspec(f2), getfullargspec(f))


    def test_empty_types(self):

        def x():
            @contract
            def f(myparam):
                """
                :param myparam: something
                """

        self.assertRaises(MissingContract, x)

    def test_empty_types2(self):

        @contract
        def f(x):
            """
            :param x: something
            :type x: *
            """

        f(1)
